@@ -1,3 +1,5 @@
+//! The OpenCL specific implementation of a [`Buffer`], [`Device`], [`Program`] and [`Kernel`].
+
 mod error;
 mod utils;
 
@@ -34,7 +36,7 @@ pub type cl_device_id = opencl3::types::cl_device_id;
 ///
 /// It is the first two identifiers of e.g. `lspci`:
 ///
-/// ```ignore
+/// ```text
 ///     4e:00.0 VGA compatible controller
 ///     || └└-- Device ID
 ///     └└-- Bus ID
@@ -202,15 +204,19 @@ impl fmt::Display for Vendor {
     }
 }
 
+/// A Buffer to be used for sending and receiving data to/from the GPU.
 pub struct Buffer<T> {
     buffer: opencl3::memory::Buffer<T>,
+    /// The number of T-sized elements.
     length: usize,
 }
 
+/// OpenCL specific device.
 #[derive(Debug, Clone)]
 pub struct Device {
     vendor: Vendor,
     name: String,
+    /// The total memory of the GPU in bytes.
     memory: u64,
     pci_id: PciId,
     uuid: Option<DeviceUuid>,
@@ -219,6 +225,9 @@ pub struct Device {
 
 impl Hash for Device {
     fn hash<H: Hasher>(&self, state: &mut H) {
+        self.vendor.hash(state);
+        self.name.hash(state);
+        self.memory.hash(state);
         self.pci_id.hash(state);
         self.uuid.hash(state);
     }
@@ -226,34 +235,44 @@ impl Hash for Device {
 
 impl PartialEq for Device {
     fn eq(&self, other: &Self) -> bool {
-        self.pci_id == other.pci_id && self.uuid == other.uuid
+        self.vendor == other.vendor
+            && self.name == other.name
+            && self.memory == other.memory
+            && self.pci_id == other.pci_id
+            && self.uuid == other.uuid
     }
 }
 
 impl Eq for Device {}
 
 impl Device {
+    /// Returns the [`Vendor`] of the GPU.
     pub fn vendor(&self) -> Vendor {
         self.vendor
     }
 
+    /// Returns the name of the GPU, e.g. "GeForce RTX 3090".
     pub fn name(&self) -> String {
         self.name.clone()
     }
+
+    /// Returns the memory of the GPU in bytes.
     pub fn memory(&self) -> u64 {
         self.memory
     }
     pub fn is_little_endian(&self) -> GPUResult<bool> {
-        match self.device.endian_little() {
-            Ok(0) => Ok(false),
-            Ok(_) => Ok(true),
-            Err(_) => Err(GPUError::DeviceInfoNotAvailable(CL_DEVICE_ENDIAN_LITTLE)),
-        }
+        self.device
+            .endian_little()
+            .map_err(|_| GPUError::DeviceInfoNotAvailable(CL_DEVICE_ENDIAN_LITTLE))
     }
+
+    /// Returns the PCI-ID of the GPU, see the [`PciId`] type for more information.
     pub fn pci_id(&self) -> PciId {
         self.pci_id
     }
 
+    /// Returns the PCI-ID of the GPU if available, see the [`DeviceUuid`] type for more
+    /// information.
     pub fn uuid(&self) -> Option<DeviceUuid> {
         self.uuid
     }
@@ -296,11 +315,20 @@ impl Device {
         utils::DEVICES.iter()
     }
 
+    /// Low-level access to the device identifier.
+    ///
+    /// It changes when the device is initialized and should only be used to interact with other
+    /// libraries that work on the lowest OpenCL level.
     pub fn cl_device_id(&self) -> cl_device_id {
         self.device.id()
     }
 }
 
+/// Abstraction that contains everything to run an OpenCL kernel on a GPU.
+///
+/// The majority of methods are the same as [`crate::cuda::Program`], so you can write code using this
+/// API, which will then work with OpenCL as well as CUDA kernels.
+#[allow(broken_intra_doc_links)]
 pub struct Program {
     device_name: String,
     queue: CommandQueue,
@@ -309,10 +337,12 @@ pub struct Program {
 }
 
 impl Program {
+    /// Returns the name of the GPU, e.g. "GeForce RTX 3090".
     pub fn device_name(&self) -> &str {
         &self.device_name
     }
 
+    /// Creates a program for a specific device from OpenCL source code.
     pub fn from_opencl(device: &Device, src: &str) -> GPUResult<Program> {
         let cached = utils::cache_path(device, src)?;
         if std::path::Path::exists(&cached) {
@@ -348,6 +378,7 @@ impl Program {
         }
     }
 
+    /// Creates a program for a specific device from a compiled OpenCL binary.
     pub fn from_binary(device: &Device, bin: Vec<u8>) -> GPUResult<Program> {
         let context = Context::from_device(&device.device)?;
         let bins = vec![&bin[..]];
@@ -374,6 +405,9 @@ impl Program {
         })
     }
 
+    /// Creates a new buffer that can be used for input/output with the GPU.
+    ///
+    /// The `length` is the number of elements to create.
     pub fn create_buffer<T>(&self, length: usize) -> GPUResult<Buffer<T>> {
         assert!(length > 0);
         let buff = opencl3::memory::Buffer::create(
@@ -414,6 +448,9 @@ impl Program {
         })
     }
 
+    /// Puts data from an existing buffer onto the GPU.
+    ///
+    /// The `offset` is in number of `T` sized elements, not in their byte size.
     pub fn write_from_buffer<T>(
         &self,
         buffer: &Buffer<T>,
@@ -422,16 +459,19 @@ impl Program {
     ) -> GPUResult<()> {
         assert!(offset + data.len() <= buffer.length, "Buffer is too small");
 
-        let buff = buffer
+        let mut buff = buffer
             .buffer
             .create_sub_buffer(CL_MEM_READ_WRITE, offset, data.len())?;
 
         self.queue
-            .enqueue_write_buffer(&buff, CL_BLOCKING, 0, data, &[])?;
+            .enqueue_write_buffer(&mut buff, CL_BLOCKING, 0, data, &[])?;
 
         Ok(())
     }
 
+    /// Reads data from the GPU into an existing buffer.
+    ///
+    /// The `offset` is in number of `T` sized elements, not in their byte size.
     pub fn read_into_buffer<T>(
         &self,
         buffer: &Buffer<T>,
@@ -462,33 +502,42 @@ impl Program {
     }
 }
 
-pub trait KernelArgument<'a> {
-    fn push(&self, kernel: &mut Kernel<'a>);
+/// Abstraction for kernel arguments.
+///
+/// The kernel doesn't support being called with custom types, hence some conversion might be
+/// needed. This trait enables automatic coversions, so that any type implementing it can be
+/// passed into a [`Kernel`].
+pub trait KernelArgument {
+    /// Apply the kernel argument to the kernel.
+    fn push(&self, kernel: &mut Kernel);
 }
 
-impl<'a, T> KernelArgument<'a> for Buffer<T> {
-    fn push(&self, kernel: &mut Kernel<'a>) {
+impl<T> KernelArgument for Buffer<T> {
+    fn push(&self, kernel: &mut Kernel) {
         kernel.builder.set_arg(&self.buffer);
     }
 }
 
-impl KernelArgument<'_> for i32 {
+impl KernelArgument for i32 {
     fn push(&self, kernel: &mut Kernel) {
         kernel.builder.set_arg(self);
     }
 }
 
-impl KernelArgument<'_> for u32 {
+impl KernelArgument for u32 {
     fn push(&self, kernel: &mut Kernel) {
         kernel.builder.set_arg(self);
     }
 }
 
+/// A local buffer.
 pub struct LocalBuffer<T> {
+    /// The number of T sized elements.
     length: usize,
     _phantom: std::marker::PhantomData<T>,
 }
 impl<T> LocalBuffer<T> {
+    /// Returns a new buffer of the specified `length`.
     pub fn new(length: usize) -> Self {
         LocalBuffer::<T> {
             length,
@@ -497,7 +546,7 @@ impl<T> LocalBuffer<T> {
     }
 }
 
-impl<T> KernelArgument<'_> for LocalBuffer<T> {
+impl<T> KernelArgument for LocalBuffer<T> {
     fn push(&self, kernel: &mut Kernel) {
         kernel
             .builder
@@ -505,6 +554,7 @@ impl<T> KernelArgument<'_> for LocalBuffer<T> {
     }
 }
 
+/// A kernel that can be executed.
 #[derive(Debug)]
 pub struct Kernel<'a> {
     builder: ExecuteKernel<'a>,
@@ -512,10 +562,13 @@ pub struct Kernel<'a> {
 }
 
 impl<'a> Kernel<'a> {
-    pub fn arg<T: KernelArgument<'a>>(mut self, t: &T) -> Self {
+    /// Set a kernel argument.
+    pub fn arg<T: KernelArgument>(mut self, t: &T) -> Self {
         t.push(&mut self);
         self
     }
+
+    /// Actually run the kernel.
     pub fn run(mut self) -> GPUResult<()> {
         self.builder.enqueue_nd_range(&self.queue)?;
         Ok(())
